/*
 * @(#)PerformanceTest.java Patrick Chan
 *
 * Copyright (c) 1994-1995 Sun Microsystems, Inc. All Rights Reserved.
 *
 * Permission to use, copy, modify, and distribute this software
 * and its documentation for NON-COMMERCIAL or COMMERCIAL purposes and
 * without fee is hereby granted. 
 * Please refer to the file http://java.sun.com/copy_trademarks.html
 * for further important copyright and trademark information and to
 * http://java.sun.com/licensing.html for further important licensing
 * information for the Java (tm) Technology.
 * 
 * SUN MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE SUITABILITY OF
 * THE SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED
 * TO THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A
 * PARTICULAR PURPOSE, OR NON-INFRINGEMENT. SUN SHALL NOT BE LIABLE FOR
 * ANY DAMAGES SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING OR
 * DISTRIBUTING THIS SOFTWARE OR ITS DERIVATIVES.
 * 
 * THIS SOFTWARE IS NOT DESIGNED OR INTENDED FOR USE OR RESALE AS ON-LINE
 * CONTROL EQUIPMENT IN HAZARDOUS ENVIRONMENTS REQUIRING FAIL-SAFE
 * PERFORMANCE, SUCH AS IN THE OPERATION OF NUCLEAR FACILITIES, AIRCRAFT
 * NAVIGATION OR COMMUNICATION SYSTEMS, AIR TRAFFIC CONTROL, DIRECT LIFE
 * SUPPORT MACHINES, OR WEAPONS SYSTEMS, IN WHICH THE FAILURE OF THE
 * SOFTWARE COULD LEAD DIRECTLY TO DEATH, PERSONAL INJURY, OR SEVERE
 * PHYSICAL OR ENVIRONMENTAL DAMAGE ("HIGH RISK ACTIVITIES").  SUN
 * SPECIFICALLY DISCLAIMS ANY EXPRESS OR IMPLIED WARRANTY OF FITNESS FOR
 * HIGH RISK ACTIVITIES.
 */

import awt.*;
import browser.*;

/**
 * To write your own performance suite, create a class that
 * extends PerformanceTest.  Next, override the 'test' method.
 * Finally, your subclass needs to call the init method
 * with an array of the names of each test.
 **/
class PerformanceTest extends Applet implements Runnable {
    /* A Test must override this method. 
       'test' specifies the test to run.  
       Returns the number operations executed.  */
    protected int test(Graphics g, int test) {
        int count = 0;
        try {
	    switch (test) {
            case -1:        // this needs to be specified
		while (!stopTest) {
		    count++;
		}
            }
	} catch (ThreadDeath e) {
        }
        return count;
    }

    /**
     * Testbox variables.
     */

    /* Location and size of test box. */
    int testBoxX, testBoxY, testBoxW, testBoxH;

    /* Font used to paint titles and results. */
    Font font;

    /* Spacing between lines. */
    int fontSp;

    /* Current test.  -1 -> no test in progress. */
    int curTest = -1;

    /* Handle of thread running a test.  null if no test is running. */
    Thread testRunning = null;

    /* Current test mode.  'a' -> all untimed ops; 'A' -> all ops;
       ' ' -> no mode. */
    char testMode = ' ';

    /* Timing duration in ms. */
    int timingDuration = 3000;

    /* update() checks this variable to see if it should clear
       the display before updating the display. */
    boolean clearScreen = false;

    /* Used by the testing thread to perform the tests. */
    Graphics testGraphics;

    /* Title used for printout. */
    String testTitle;

    /* Title for each test. */
    String testLabels[];

    /* Cost of null loop in milliseconds.  i.e. while(true) int++ */
    float nullloopMs = 0.0;

    /* If true, the test must stop. */
    boolean stopTest;

    /* The raw time(ms) per operation, before subtraction of the
       null loop overhead.  */
    float testResults[];

    /**
     * Initialize the tests.  
     * 't' is the testTitle; it's used when printing the test results.
     * 'tl' is the list of test labels; the size of this array
     * specifies the number of tests.
     */
    public void init(String t, String tl[]) {
        int w;

        testTitle = t;
        testLabels = tl;
        testResults = new float[testLabels.length];
        for (int i=0; i<testResults.length; i++) {
            testResults[i] = -1.0;
        }

        font = getFont("Helvetica", Font.PLAIN, 14);
        fontSp = 2;

        // Setup test box.
        testBoxX = 0;
        testBoxY = 0;
        testBoxH = (font.height + fontSp) * testLabels.length;
        testBoxW = 0;
        for (int i=0; i<testLabels.length; i++) {
	    w = font.stringWidth(testLabels[i] + ": ");
	    if (w > testBoxW) {
		testBoxW = w;
	    }
	}
        resize(550, testBoxY + testBoxH + font.height * 2);
    }

    protected void start() {
	getFocus();
    }

    /**
     * Paint the screen.
     */
    public void paint(Graphics g) {
        g.clearRect(0, 0, width, height);
        update(g);
    }

    public void update(Graphics g) {
        if (clearScreen) {
            g.clearRect(0, 0, width, height);
            clearScreen = false;
        }
	int y = testBoxY + font.height;

	g.setFont(font);
	// Paint titles
	for (int i=0; i<testLabels.length; i++) {
	    g.setForeground(Color.black);
            if (i == curTest) {
	        g.setForeground(Color.red);
            }
	    g.drawString(testLabels[i] + ": ", testBoxX, y);
	    y += font.height + fontSp;
	}

	// Paint results.
	g.setForeground(Color.blue);
	y = testBoxY + font.height;
	for (int i=0; i<testLabels.length; i++) {
	    g.drawString(printTestResult(i), testBoxW + 10, y);
	    y += font.height + fontSp;
	}

        // Paint info line.
	Font f = getFont("Helvetica", Font.PLAIN, 12);
	String s = 
	    "Timing duration: "+(timingDuration/1000)+"s   "
	    + "Loop overhead: "+(nullloopMs*1000)+"us";
	int w = f.stringWidth(s);
	g.setForeground(Color.black);
	g.setFont(f);
	g.drawString(s, width-w, height - f.height);

        // Start test.
        if (curTest >= 0 && isTestRunning() == null) {
            // is it possible to have more than one update come in?
            // if so, there's a race condition here.
            testGraphics = new Graphics(g.drawSurface, 
		g.originX, g.originY, g.scaleX, g.scaleY);
            Thread t = new Thread(this);
            setTestRunning(t);
            t.start();
        } 
    }

    /**
     * keyDown
     */
    public void keyDown(int key) {
	char ch = (char)key;

	switch (ch) {
        case '0':
            if (isTestRunning() == null) {
		computeNullloop(null);
		clearScreen = true;
            }
            break;
        case 's':
	    testMode = ' ';
            stopTest = true;
            break;
	case 'p':
            if (isTestRunning() == null) {
		System.out.println(testTitle);
		for (int i=0; i<testTitle.length(); i++) {
		    System.out.print("-");
		}
		System.out.println();

                // Determine max width of labels.
                int w = 0;
		for (int i=0; i<testLabels.length; i++) {
                    w = Math.max(w, testLabels[i].length());
                }

                // Now print labels with results.
		for (int i=0; i<testLabels.length; i++) {
		    System.out.print(testLabels[i]+":  ");
		    for (int j=w-testLabels[i].length(); j>=0; j--) {
			System.out.print(" ");
		    }
		    System.out.print(printTestResult(i));
		    System.out.println();
		}
	    }
	    break;
	case 'a':
	    if (isTestRunning() == null && testMode == ' ') {
		// Skip completed tests.
		curTest = 0;
		while(curTest < testResults.length 
		    && testResults[curTest] >= 0.0) {
		    curTest++;
		}
                if (curTest == testResults.length) {
                    curTest = -1;
                } else {
		    testMode = 'a';
                }
	    }
	    break;
	case 'A':
	    if (isTestRunning() == null && testMode == ' ') {
		testMode = 'A';
		curTest = 0;
	    }
	    break;
	case '+':
            if (isTestRunning() == null) {
		if (timingDuration < 5000) {
		    timingDuration += 1000;
		} else {
		    timingDuration += 5000;
		}
		clearScreen = true;
            }
	    break;
	case '-':
            if (isTestRunning() == null) {
		if (timingDuration > 5000) {
		    timingDuration -= 5000;
		} else if (timingDuration > 1000) {
		    timingDuration -= 1000;
		}
		clearScreen = true;
            }
	    break;
	default:
	    testMode = ' ';
	}
        repaint();
    }

    /**
     * mouseDown
     */
    public void mouseDown(int x, int y) {
        int sel;
        
	testMode = ' ';
        if (isTestRunning() != null) {
	    stopTest = true;
        } else {
	    y -= testBoxY;
	    sel = y / (font.height + fontSp);
	    sel = Math.min(sel, testLabels.length - 1);
	    curTest = sel;
        }
        clearScreen = true;
	repaint();
    }

    /* Computes the cost of the null loop. */
    void computeNullloop(Thread testThread) {
	// Need to compute cost of null loop.
	PerformanceTestTimer.start(this, testThread, 2000, null);
	double start = System.nowMillis();
	int count = test(null, -1);
	nullloopMs = (float)(System.nowMillis() - start) / count;
    }

    public void run() {
	int count = 0;
	double start = 0; 
	double time = 0;

        if (nullloopMs == 0.0) {
            computeNullloop(Thread.currentThread());
        }

	PerformanceTestTimer.start(this, Thread.currentThread(), 
	    timingDuration, testGraphics);
	start = System.nowMillis();
	count = test(testGraphics, curTest);
	time = (double)System.nowMillis() - start;
	testGraphics.dispose();
        testResults[curTest] = (float)(time / count);

	if (testMode == 'a' && curTest < testLabels.length - 1) {
	    curTest++;
	    while(curTest < testResults.length 
		&& testResults[curTest] > 0.0) {
		curTest++;
	    }
	} else if (testMode == 'A' && curTest < testLabels.length - 1) {
            curTest++;
	} else {
	    curTest = -1;
	    testMode = ' ';
	}
        setTestRunning(null);
        clearScreen = true;
        repaint();
    }

    String printTestResult(int t) {
        String result;
        float msPerOp = testResults[t] - nullloopMs;
        float f;

        f = round2Places(1000.0 / msPerOp);
        if (f < 0.0) {
            return "";
        }
        if (f > 1000000) {
	    result = (f / 1000000) + "M";
        } else if (f > 1000) {
	    result = (f / 1000) + "K";
        } else {
	    result = "" + f;
        }
        result += " ops/s   ";

        f = round2Places(msPerOp);
        if (f > 0.0 && f < 1) {
	    result += (f * 1000) + " u";
        } else if (f > 1000) {
	    result += (f / 1000) + " ";
        } else {
	    result += f + " m";
        }
        result += "s/op";
        return result;
    }

    float round2Places(double d0) {
        // I think there's a bug with fp on the PC.
        float d = (float)d0;
	float mult = 1.0;

	if (d > 0.0 && d < 100.0) {
	    while (d < 10.0) {
		d *= 10.0;
		mult /= 10.0;
	    }
	} else if (d > 0.0) {
	    while (d > 0.0 && d > 100.0) {
		d /= 10.0;
		mult *= 10.0;
	    }
	}
	return (float)Math.round(d) * (float)mult;
    }

    synchronized Thread isTestRunning() {
        return testRunning;
    }
 
    synchronized void setTestRunning(Thread t) {
        testRunning = t;
    }
}

class PerformanceTestTimer extends Thread {
    /* Handle to the test thread. */
    Thread testThread;

    /* Number of milliseconds before stopping the test thread. */
    int timingDuration;

    /* Copy of the applet's graphics context. */
    Graphics graphics;
    
    /* Handle to the Applet. */
    PerformanceTest applet;

    /* Makes a copy of Graphics g. */
    public static void start(PerformanceTest a, Thread t, int d, Graphics g) {
        PerformanceTestTimer self = new PerformanceTestTimer();

        self.applet = a;
        self.applet.stopTest = false;
        self.testThread = t;
	self.timingDuration  = d;
        if (g != null) {
	    self.graphics = new Graphics(g.drawSurface, 
		g.originX, g.originY, g.scaleX, g.scaleY);
        }
	self.setPriority(Thread.MAX_PRIORITY);
	self.start();
    }
    
    public void run() {
        Font font = applet.getFont("Courier", Font.BOLD, 36);
        int charW = 3 * font.widths['M'];
        int width = applet.width;
        double start, end;

        if (graphics != null) {
	    graphics.setFont(font);
	    graphics.setForeground(Color.red);
        }
        while (applet.isTestRunning() == testThread
                && !applet.stopTest 
                && timingDuration > 0) {
	    if (graphics != null) {
		graphics.clearRect(width-charW, 0, charW, font.height);
		graphics.drawString(""+(Math.ceil((float)timingDuration / 1000.0)), 
		    width-charW, font.height);
                graphics.wServer.sync();
	    }
            start = System.nowMillis();
	    sleep(1000);
            end = System.nowMillis();
            
            timingDuration -= (int)(end - start);
	}
        if (applet.isTestRunning() == testThread) {
	    applet.stopTest = true;
        }
        if (graphics != null) {
	    graphics.dispose();
        }
    }
}
 
